/* Important Acknowledgements: */
/* --------------------------- */
/*	 The majority of this pi calculation 'engine' comes from	*/
/*  a great little command line app written by Roy Williams.	*/
/*  I found his app a few years ago, and wanted to figure out   */
/*	how it worked.  I ended up making a nice GUI for it that	*/
/*  would let you keep track of the calculation's progress,		*/
/*  time remaining, etc.  That's how this application came		*/
/*  about.														*/
/*																*/
/*  Roy's web site can be found at:								*/
/*				http://www.ccsf.caltech.edu/~roy/				*/
/*
/*	More fun info/links about pi (as well as the original code	*/
/*  I used to create this app) can be found at his pi page:		*/
/*				http://www.ccsf.caltech.edu/~roy/upi/pi.html	*/
/*																*/


#include "stdafx.h"

#include <string.h>
#include <stdio.h>
#include <time.h>
#include <math.h>

#define BASE 10000L

long *tot;
long *t1;
long *t2;
long *t3;
long *t4;
long *t5;

long nblock;
long digits_per_block = 1;

size_t wordsize;
volatile bool stop_flag = false;
volatile long total_num_passes;
volatile long current_pass_thread_A;
volatile long current_pass_thread_B;

void copy(long *result, long *from)
{
	memcpy (result, from, wordsize);
}

int zero(long *result)
{
	long i;
	for(i=0; i<nblock; i++)
		if(result[i])
			return 0;
	return 1;
}

void add(long *result, long *increm)
{
	long i;
	for(i=nblock-1; i>=0; i--){
		result[i] += increm[i];
		if(result[i] >= BASE){
			result[i] -= BASE;
			result[i-1]++;
		}
	}
}

void sub(long *result, long *decrem)
{
	long i;
	for(i=nblock-1; i>=0; i--){
		result[i] -= decrem[i];
		if(result[i] < 0){
			result[i] += BASE;
			result[i-1]--;
		}
	}
}

void mult(long *result, long factor)
{
	long i, carry = 0;
	for(i=nblock-1; i>=0; i--){
		result[i] *= factor;
		result[i] += carry;
		carry = result[i]/BASE;
		result[i] %= BASE;
	}
}

void div(long *result, long denom)
{
	long i, carry = 0;

	for(i=0; i<nblock; i++){
		result[i] += carry*BASE;
		carry = result[i] % denom;
		result[i] /= denom;
	}
}

void set(long *result, long rhs)
{
	long i;
	for(i=0; i<nblock; i++)
		result[i] = 0;
	result[0] = rhs;
}

void FormatResult(long *result, CString &strResult, FILE *f, long digits_desired)
{
	long i, k;
	char s[100];
	long strPrinted = 0;
	long fPrinted = 0;

	long display_digits = digits_desired;
	if (display_digits > 20000)
		display_digits = 20000;  // CStrings can't handle >32k on Win 9x

	// begin the output
	strResult.Format("%1d.\r\n", result[0]);
	if (f != NULL)
		fprintf(f, "%1d.\n", result[0]);

	for(i=1; i<nblock; i++)
	{
	  sprintf(s, "%0*d", digits_per_block, result[i]);
	  for (k = 0; k < digits_per_block; k++)
	  {
		// print to the file
		if ((fPrinted < digits_desired) && (f != NULL))
		{
			fputc((char)s[k], f);
			fPrinted++;
		    if((fPrinted % 5) == 0) fputc(' ', f);
			if((fPrinted % 50) == 0) fputc('\n', f);
			if((fPrinted % 1000) == 0) fputc('\n', f);
		}

		// print to CString
		if (strPrinted < display_digits)
		{
			strResult += s[k];
			strPrinted++;
		    if((strPrinted % 5) == 0) strResult += " ";
			if((strPrinted % 50) == 0) strResult += "\r\n";
			if((strPrinted % 1000) == 0) strResult += "\r\n";
		}
	  }
	}

	if (display_digits != digits_desired)
	{
		strResult += 
			"\r\nNote: Due to MS Windows limitations, this window can "\
			"display only the first 20,000 digits of pi. "\
			"However, the entire results were saved in the file 'pi_digits.txt'.";
	}
}

void arctan(long *result, long *w1, long *w2, long denom, long onestep, volatile long &pass_counter)
{
	long denom2 = denom*denom;
	long k = 1;

	set(result, 1);
	div(result, denom);
	copy(w1, result);

	do
	{
		if(onestep)
		{
			div(w1, denom2);
		}
		else 
		{
			div(w1, denom);
			div(w1, denom);
		}
		copy(w2, w1);
		div(w2, 2*k+1);
		if(k%2)
			sub(result, w2);
		else
			add(result, w2);
		k++;
		pass_counter++;
	} while(!zero(w2) && (stop_flag == false));
}

volatile bool Term2ThreadComplete = false;
UINT Term2Thread(LPVOID param)
{
	// compute the second term
	arctan(t3, t4, t5, 239, 1, current_pass_thread_B);

	if (!stop_flag)
		current_pass_thread_B = nblock;

	Term2ThreadComplete = true;
	return (0);
}

bool calculate_pi(long int digits_desired, CString &Result, bool bSaveToFile)
{
	digits_per_block = static_cast<long>(log10(BASE)); 
	fprintf(stderr,"Using %ld digits/block\n", digits_per_block);

	// so that the last few digits are always correct...
	long int ndigit = digits_desired + 10; 
	if(ndigit < 20) ndigit = 20;

	nblock = ndigit/digits_per_block;
	wordsize = nblock*sizeof(long);

	// make sure the machine has enough *physical* memory,
	// otherwise, this thread may be waaaay to slow
	MEMORYSTATUS MemStat;
	MemStat.dwLength = sizeof(MemStat);
	GlobalMemoryStatus(&MemStat);
	DWORD MemRequired = 6L * nblock * sizeof(long);
	MemRequired = MemRequired * 3 / 2;	// unrealistic to use more than 66% of mem
	if (MemRequired > MemStat.dwTotalPhys)
	{
		Result.Format("Calculating %ld digits requires %ldM of memory,\r\n"\
			"but this machine has only %ldM.  Try fewer digits.", 
			digits_desired,
			MemRequired / 1024 / 1024,
			MemStat.dwTotalPhys / 1024 / 1024);
		return (false);
	}

	// create the file
	FILE *f = NULL;
	if (bSaveToFile)
	{
		f = fopen("pi_digits.txt", "w");
		if (f == NULL)
		{
			Result.Format("Could not create output file, pi_digits.txt.");
			return (false);
		}
	}

	// get the mem
	tot = (long *)malloc(nblock*sizeof(long));
	t1 = (long *)malloc(nblock*sizeof(long));
	t2 = (long *)malloc(nblock*sizeof(long));
	t3 = (long *)malloc(nblock*sizeof(long));
	t4 = (long *)malloc(nblock*sizeof(long));
	t5 = (long *)malloc(nblock*sizeof(long));

	if(!tot || !t1 || !t2 || !t3)
	{
		// free the mem
		if (t1) free(t1);
		if (t2) free(t2);
		if (t3) free(t3);
		if (t4) free(t4);
		if (t5) free(t5);
		if (tot) free(tot);

		// close the file
		if (f != NULL)
			fclose(f);

		Result.Format("Not enough memory... try fewer digits.");
		return (false);
	}

	// set progress to zero
	current_pass_thread_A = 0;
	current_pass_thread_B = 0;

	// found emperically...
	total_num_passes = 4 * nblock;

	// start a separate thread to compute the arctan(1/239) term.
	// This will allow dual-processor machines to be used fully
	Term2ThreadComplete = false;
	if (AfxBeginThread(Term2Thread, NULL, THREAD_PRIORITY_NORMAL) == NULL)
		stop_flag = true;

	arctan(tot, t1, t2, 5, 0, current_pass_thread_A);
	mult(tot, 4);

	// should be about here after pass one
	if (!stop_flag)
		current_pass_thread_A = 3 * nblock;

	// Wait until the Term2Thread finishes
	while (Term2ThreadComplete == false)
	{
		Sleep(100);
	}

	if (stop_flag == false)
	{
		// complete the calculations
		sub(tot, t3);
		mult(tot, 4);

		// output the results
		FormatResult(tot, Result, f, digits_desired);
	}

	// free the memory
	free (tot);
	free (t1);
	free (t2);
	free (t3);
	free (t4);
	free (t5);

	// close the file
	if (f != NULL)
		fclose(f);

	return (!stop_flag);
}