// picalcDlg.cpp : implementation file
//

#include "stdafx.h"
#include "math.h"
#include "picalc.h"
#include "picalcDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

// External Stuff
bool calculate_pi(long int digits_desired, CString &Result, bool bSaveToFile);
extern volatile bool stop_flag;
extern volatile long total_num_passes;
extern volatile long current_pass_thread_A;
extern volatile long current_pass_thread_B;


/////////////////////////////////////////////////////////////////////////////
// CAboutDlg dialog used for App About

class CAboutDlg : public CDialog
{
public:
	CAboutDlg();

// Dialog Data
	//{{AFX_DATA(CAboutDlg)
	enum { IDD = IDD_ABOUTBOX };
	CString	m_Info;
	//}}AFX_DATA

	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(CAboutDlg)
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	//}}AFX_VIRTUAL

// Implementation
protected:
	//{{AFX_MSG(CAboutDlg)
	virtual BOOL OnInitDialog();
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()
};

CAboutDlg::CAboutDlg() : CDialog(CAboutDlg::IDD)
{
	//{{AFX_DATA_INIT(CAboutDlg)
	m_Info = _T("");
	//}}AFX_DATA_INIT
}

void CAboutDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CAboutDlg)
	DDX_Text(pDX, IDC_INFO, m_Info);
	//}}AFX_DATA_MAP
}

BOOL CAboutDlg::OnInitDialog() 
{
	CDialog::OnInitDialog();

	// "About" box info
	m_Info.Format(_TEXT(
		"This is a neat little program that can calculate pi to any (reasonable) precision. "\
		"It's pretty simple to use:  Just enter the number of digits you want to calculate, "\
		"then click the 'Go!' button.  It's useful up to a half million digits or so, after which "\
		"you'll be waiting a lonnnng time.\r\n\r\n"\
		
		"It uses Machin's formula, pi/4 = 4 arctan(1/5) - arctan(1/239), "\
		"combined with the Gregory series for computing arctangents.  The formula "\
		"converges reasonably quickly, and is relatively easy to code.\r\n\r\n"\

		"Important Acknowledgements:\r\n"\
		"------------------------------------\r\n"\
		"The majority of this pi calculation 'engine' comes from\r\n"\
		"a great little command line app written by Roy Williams.\r\n"\
		"I found his app a few years ago, and wanted to figure out\r\n"\
		"how it worked.  I ended up making a nice GUI for it that\r\n"\
		"would let you keep track of the calculation's progress,\r\n"\
		"time remaining, etc.  That's how this application came\r\n"\
		"about.\r\n"\
		"\r\n"\
		"Roy's web site can be found at:\r\n"\
		"   http://www.ccsf.caltech.edu/~roy/\r\n"\
		"\r\n"\
		"More fun info/links about pi (as well as the original code\r\n"\
		"I used to create this app) can be found at his pi page:\r\n"\
		"   http://www.ccsf.caltech.edu/~roy/upi/pi.html\r\n\r\n"\

		"In case you haven't guessed, I've been fascinated with pi for many years. "\
		"I've written a couple of other programs to calculate pi.  The first one I wrote "\
		"(many years ago) was quite slow: It took an hour or so to calculate only 6 digits. "\
		"I've written a couple of other ones since then, but this is the first time I've "\
		"packaged it up in a friendly, easy to use application.\r\n\r\n"\

		"I enjoyed writing this app, and I hope you have fun playing with it.  This program is "\
		"freeware... feel free to use or distribute it.  All I ask is that you drop me an "\
		"email to tell me what you think of it.   Also, feel free to email me with suggestions "\
		"or <<gasp>> bug reports.  If you *really* enjoy it, and you happen "\
		"to know me, drop me a Snapple iced tea the next time you see me... it's my favorite "\
		"nectar for those late night programming sessions, and it'll bring a nice big smile "\
		"to my face :-)\r\n\r\n"\

		"Well, it is now 5:30 am and I really should get some sleep.  Happy calculating!\r\n\r\n"\

		"copyright (c) 2003 Eric 'Pi' Minbiole\r\n"\
		"Email: eric@pisymbol.com\r\n"\
		"Web: www.pisymbol.com\r\n"\
		
		));

	UpdateData(FALSE);
	
	return TRUE;  // return TRUE unless you set the focus to a control
	              // EXCEPTION: OCX Property Pages should return FALSE
}

BEGIN_MESSAGE_MAP(CAboutDlg, CDialog)
	//{{AFX_MSG_MAP(CAboutDlg)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CPicalcDlg dialog

CPicalcDlg::CPicalcDlg(CWnd* pParent /*=NULL*/)
	: CDialog(CPicalcDlg::IDD, pParent)
{
	//{{AFX_DATA_INIT(CPicalcDlg)
	m_NumDigits = 0;
	m_Results = _T("");
	m_Status = _T("");
	//}}AFX_DATA_INIT
	// Note that LoadIcon does not require a subsequent DestroyIcon in Win32
	m_hIcon = AfxGetApp()->LoadIcon(IDR_MAINFRAME);
}

void CPicalcDlg::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(CPicalcDlg)
	DDX_Control(pDX, IDC_PROGRESS1, m_Progress);
	DDX_Text(pDX, IDC_NUM_DIGITS, m_NumDigits);
	DDX_Text(pDX, IDC_RESULTS, m_Results);
	DDX_Text(pDX, IDC_STATUS, m_Status);
	//}}AFX_DATA_MAP
}

BEGIN_MESSAGE_MAP(CPicalcDlg, CDialog)
	//{{AFX_MSG_MAP(CPicalcDlg)
	ON_WM_SYSCOMMAND()
	ON_WM_PAINT()
	ON_WM_QUERYDRAGICON()
	ON_BN_CLICKED(IDGO, OnGo)
	ON_WM_TIMER()
	ON_BN_CLICKED(IDC_ABOUT, OnAbout)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CPicalcDlg message handlers

BOOL CPicalcDlg::OnInitDialog()
{
	CDialog::OnInitDialog();

	// Add "About..." menu item to system menu.

	// IDM_ABOUTBOX must be in the system command range.
	ASSERT((IDM_ABOUTBOX & 0xFFF0) == IDM_ABOUTBOX);
	ASSERT(IDM_ABOUTBOX < 0xF000);

	CMenu* pSysMenu = GetSystemMenu(FALSE);
	if (pSysMenu != NULL)
	{
		CString strAboutMenu;
		strAboutMenu.LoadString(IDS_ABOUTBOX);
		if (!strAboutMenu.IsEmpty())
		{
			pSysMenu->AppendMenu(MF_SEPARATOR);
			pSysMenu->AppendMenu(MF_STRING, IDM_ABOUTBOX, strAboutMenu);
		}
	}

	// Set the icon for this dialog.  The framework does this automatically
	//  when the application's main window is not a dialog
	SetIcon(m_hIcon, TRUE);			// Set big icon
	SetIcon(m_hIcon, FALSE);		// Set small icon
	
	// TODO: Add extra initialization here
	m_NumDigits = 10;
	m_Status.Format(_TEXT(""));
	UpdateData(FALSE);

	// do a quick timing test to gauge the speed of this computer
	CString dummy;
	m_NumDigitsInInitialCalculation = 100;
	m_TimeToDoInitialCalculation = 1;
	int iterations = 0;

	while ((m_TimeToDoInitialCalculation < 200) && (iterations < 10))
	{
		// try more and more digits each time
		m_NumDigitsInInitialCalculation *= 2;

		// do a timing test
		DWORD StartTime = GetTickCount();
		if (!calculate_pi(m_NumDigitsInInitialCalculation, dummy, false))
			break;	// failed!
		DWORD EndTime = GetTickCount();
		m_TimeToDoInitialCalculation = EndTime - StartTime;

		iterations++;
	}

	m_State = STATE_INACTIVE;
	m_TimerId = SetTimer(1234, 100, NULL);

	if (m_TimerId == NULL)
	{
		m_Results.Format(_TEXT("Error: Could not create timer.  Pi Calc will not work."));
		GetDlgItem(IDGO)->EnableWindow(FALSE);
		UpdateData(FALSE);
	}
	
	return TRUE;  // return TRUE  unless you set the focus to a control
}

void CPicalcDlg::OnSysCommand(UINT nID, LPARAM lParam)
{
	if ((nID & 0xFFF0) == IDM_ABOUTBOX)
	{
		CAboutDlg dlgAbout;
		dlgAbout.DoModal();
	}
	else
	{
		CDialog::OnSysCommand(nID, lParam);
	}
}

// If you add a minimize button to your dialog, you will need the code below
//  to draw the icon.  For MFC applications using the document/view model,
//  this is automatically done for you by the framework.

void CPicalcDlg::OnPaint() 
{
	if (IsIconic())
	{
		CPaintDC dc(this); // device context for painting

		SendMessage(WM_ICONERASEBKGND, (WPARAM) dc.GetSafeHdc(), 0);

		// Center icon in client rectangle
		int cxIcon = GetSystemMetrics(SM_CXICON);
		int cyIcon = GetSystemMetrics(SM_CYICON);
		CRect rect;
		GetClientRect(&rect);
		int x = (rect.Width() - cxIcon + 1) / 2;
		int y = (rect.Height() - cyIcon + 1) / 2;

		// Draw the icon
		dc.DrawIcon(x, y, m_hIcon);
	}
	else
	{
		CDialog::OnPaint();
	}
}

// The system calls this to obtain the cursor to display while the user drags
//  the minimized window.
HCURSOR CPicalcDlg::OnQueryDragIcon()
{
	return (HCURSOR) m_hIcon;
}

void CPicalcDlg::EnableControls(BOOL Enable) 
{
	GetDlgItem(IDC_NUM_DIGITS)->EnableWindow(Enable);
	GetDlgItem(IDC_RESULTS)->EnableWindow(Enable);

	if (Enable)
		GetDlgItem(IDGO)->SetWindowText(_TEXT("Go!"));
	else
		GetDlgItem(IDGO)->SetWindowText(_TEXT("Stop"));
}


const DWORD PRE_DELAY = 200;
UINT CPicalcDlg::CalculationThread(LPVOID param)
{
	// get a pointer to the dialog
	CPicalcDlg *pDlg = static_cast<CPicalcDlg *>(param);

	// give the UI thread some time to draw stuff
	// plus, it makes it look like we're doing stuff :-)
	Sleep(PRE_DELAY);

	// do the calculation
	CString temp;
	bool result = calculate_pi(pDlg->m_NumDigits, temp, true);
	pDlg->m_Results = temp;

	// signal the UI thread that we've finished
	if (result == true)
		pDlg->m_State = STATE_FINISHED;
	else
		pDlg->m_State = STATE_FAILED;

	return (0);
}

void CPicalcDlg::OnGo() 
{
	// start if we're inactive, stop otherwise
	if (m_State == STATE_INACTIVE)
	{
		UpdateData(TRUE);

		// do some validation of the nuber of digits
		if (m_NumDigits < 0)
		{
			CString ErrorMessage;
			ErrorMessage.Format(_TEXT("I can't even fathom how to calculate a negative numbers of digits."));
			AfxMessageBox(ErrorMessage);
			return;
		}
		else if (m_NumDigits >= 100000000)
		{
			CString ErrorMessage;
			ErrorMessage.Format(_TEXT("%ld million digits??  We won't be alive that long.\r\nTry something a bit smaller."), m_NumDigits / 1000000);
			AfxMessageBox(ErrorMessage);
			return;
		}

		// time for calculation is proportional to Digits^2
		double DigitRatio = (double)m_NumDigits / m_NumDigitsInInitialCalculation;
		double log2DigitRatio = log(DigitRatio) / log(2);
		m_ApproxTimeInSecs = m_TimeToDoInitialCalculation * pow(4, log2DigitRatio);
		m_ApproxTimeInSecs /= 1000;

		// clear the results box
		m_Results.Empty();

		// disable the controls
		EnableControls(FALSE);

		// set up the shared vars between UI and calc threads
		stop_flag = false;
		current_pass_thread_A = 0;
		current_pass_thread_B = 0;
		total_num_passes = 0;
		m_StartTime = GetTickCount();
		m_State = STATE_CALCULATING;

		// redraw stuff immediately upon button press
		OnTimer(0);

		// start up the calculation threads
		AfxBeginThread(CalculationThread, this, THREAD_PRIORITY_NORMAL);
	}
	else
	{
		stop_flag = true;
	}
}

// self explanatory
CString FormatTime(double Seconds, int NumPlaces)
{
	CString TimeString;
	if (Seconds < 60)
	{
		TimeString.Format(_TEXT("%.*f seconds"), NumPlaces, Seconds);
	}
	else if (Seconds > 3600.0 * 24 * 365)
	{
		double YearsRemaining = Seconds / (3600.0 * 24 * 365);

		TimeString.Format(_TEXT("%.2f years"), YearsRemaining);
	}
	else if (Seconds > 3600L * 24)
	{
		double DaysRemaining = Seconds / (3600L * 24);

		TimeString.Format(_TEXT("%.2f days"), DaysRemaining);
	}
	else
	{
		long SecondsRemaining = (long)Seconds;

		long Hours = SecondsRemaining / 3600;
		SecondsRemaining %= 3600;

		long Minutes = SecondsRemaining / 60;
		SecondsRemaining %= 60;

		TimeString.Format(_TEXT("%02ld:%02ld:%02ld"), Hours, Minutes, SecondsRemaining);
	}

	return (TimeString);
}


// updates the progress bar, time remaining, etc a few times per second
void CPicalcDlg::OnTimer(UINT nIDEvent) 
{
	static DWORD time_of_last_update = 0;
	static int prev_percentage_complete = -1;
	static double Rate = 1;
	static long lastCalculatedTimeRemaining = 0;

	// determine how long we've been running
	double ElapsedTime = (GetTickCount() - m_StartTime - PRE_DELAY) / 1000.0;
	if (ElapsedTime < 0)
		ElapsedTime = 0;

	// don't need to do anything unless we're active
	if (m_State != STATE_INACTIVE)
	{
		// calculate how far along we are
		int percentage_complete = 0;
		if (total_num_passes > 0)
		{
			long current_pass = current_pass_thread_A + current_pass_thread_B;
			percentage_complete = 100 * current_pass / total_num_passes;
		}
		m_Progress.SetPos(percentage_complete);

		// format the status line based on the state
		switch (m_State)
		{
		case STATE_CALCULATING:
			if (stop_flag)
			{
				// we're trying to kill the calc thread
				m_Status.Format(_TEXT("Aborting calculations..."));
			}
			else if (percentage_complete >= 100)
			{
				// calc thread almost done
				m_Status.Format(_TEXT("Formatting result..."));
			}
			else if ((percentage_complete > 0) && (ElapsedTime > 0) && (m_ApproxTimeInSecs > 5))
			{
				// calculations are under way.  Display the approx time remaining
				static long timeRemaining = 0;
				if (percentage_complete != prev_percentage_complete)
				{
					// update the rate only when the percentage changes
					Rate = percentage_complete / ElapsedTime;
					prev_percentage_complete = percentage_complete;

					lastCalculatedTimeRemaining = (long)((100 - percentage_complete) / Rate);
					timeRemaining = lastCalculatedTimeRemaining;
					time_of_last_update = GetTickCount();
				}
				else
				{
					// count down from last calculated
					timeRemaining = lastCalculatedTimeRemaining - 
						(GetTickCount() - time_of_last_update) / 1000;
				}
				if (timeRemaining < 0)
					timeRemaining = 0;
				m_Status.Format(_TEXT("Calculating...            (Approx Time Remaining: %s)"),
					FormatTime(timeRemaining, 0));
			}
			else
			{
				// calculations aren't under way because the number of digits is huge
				// or calcs will be too quick to bother displaying time remaining
				if (m_ApproxTimeInSecs > 200)
				{
					// it's going to take a while... show 'em the estimate
					m_Status.Format(_TEXT("Calculating...            (This will take approx %s)"),
						FormatTime(m_ApproxTimeInSecs, 1));
				}
				else
				{
					// won't take that long... don't bother estimating
					m_Status.Format(_TEXT("Calculating..."));
				}
			}
			break;

		case STATE_FAILED:
			if (stop_flag)
			{
				m_Status.Format(_TEXT("Calculation aborted"));
			}
			else
			{
				m_Status.Format(_TEXT("Calculation failed"));
			}
			m_State = STATE_INACTIVE;
			break;

		case STATE_FINISHED:
			m_Status.Format(_TEXT("Calculated %lu digits of pi in %s."), 
				m_NumDigits, FormatTime(ElapsedTime, 1));
			// recalculate computer speed if appropriate
			if (m_NumDigits > m_NumDigitsInInitialCalculation)
			{
				m_TimeToDoInitialCalculation = static_cast<DWORD>(ElapsedTime * 1000 + 1);
				m_NumDigitsInInitialCalculation = m_NumDigits;
			}
			m_State = STATE_INACTIVE;
			break;
		}

		if (m_State == STATE_INACTIVE)
		{
			// re-enable the controls
			EnableControls(TRUE);

			// reset vars
			prev_percentage_complete = -1;
		}

		UpdateData(FALSE);
	}

	CDialog::OnTimer(nIDEvent);
}

bool AboutShown = false;
void CPicalcDlg::OnCancel() 
{
	// tell the calc the thread to kill itself
	stop_flag = true;

	// wait for it to die
	while (m_State == STATE_CALCULATING)
	{
		Sleep(10);
	}

	// if I'm being mean, make 'em look at the "about" box at least once
//	if (!AboutShown)
//		OnAbout();

	if (m_TimerId)
		KillTimer(m_TimerId);

	CDialog::OnCancel();
}

// display fascinating stuff about Pi Calc
void CPicalcDlg::OnAbout() 
{
	AboutShown = true;
	CAboutDlg dlgAbout;
	dlgAbout.DoModal();
}
